var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
import launchEditor from "launch-editor";
import { readFileSync, createWriteStream, readFile } from "node:fs";
import { createServer, get } from "node:http";
import * as path from "node:path";
import { resolve, extname, join } from "node:path";
import { createGunzip, createInflate } from "node:zlib";
import * as os from "node:os";
import * as SourceMap from "source-map";
const DEFAULT_PORT = 8969;
const SERVER_IDENTIFIER = "spotlight-by-sentry";
const CONTEXT_LINES_ENDPOINT = "/contextlines";
async function getGeneratedCodeFromServer(filename) {
  try {
    const generatedCodeResponse = await fetch(filename);
    return generatedCodeResponse.text();
  } catch {
    return void 0;
  }
}
function parseStackTrace(requestBody) {
  try {
    return JSON.parse(requestBody);
  } catch {
    return void 0;
  }
}
async function applySourceContextToFrame(sourceMapContent, frame) {
  const consumer = await new SourceMap.SourceMapConsumer(sourceMapContent);
  const originalPosition = consumer.originalPositionFor({
    line: frame.lineno,
    column: frame.colno,
    bias: SourceMap.SourceMapConsumer.LEAST_UPPER_BOUND
  });
  if (originalPosition.source && originalPosition.line && originalPosition.column) {
    frame.lineno = originalPosition.line;
    frame.colno = originalPosition.column;
    const filePath = new URL(frame.filename).pathname.slice(1);
    frame.filename = path.resolve(path.join(path.dirname(filePath), originalPosition.source));
    const content = consumer.sourceContentFor(originalPosition.source);
    const lines = (content == null ? void 0 : content.split(os.EOL)) ?? [];
    addContextLinesToFrame(lines, frame);
  }
  return originalPosition;
}
function addContextLinesToFrame(lines, frame, linesOfContext = 5) {
  const maxLines = lines.length;
  const sourceLine = Math.max(Math.min(maxLines - 1, frame.lineno - 1), 0);
  frame.pre_context = lines.slice(Math.max(0, sourceLine - linesOfContext), sourceLine).map((line) => snipLine(line, 0));
  frame.context_line = snipLine(lines[Math.min(maxLines - 1, sourceLine)], frame.colno || 0);
  frame.post_context = lines.slice(Math.min(sourceLine + 1, maxLines), sourceLine + 1 + linesOfContext).map((line) => snipLine(line, 0));
}
function snipLine(line, colno) {
  let newLine = line;
  const lineLength = newLine.length;
  if (lineLength <= 150) {
    return newLine;
  }
  if (colno > lineLength) {
    colno = lineLength;
  }
  let start = Math.max(colno - 60, 0);
  if (start < 5) {
    start = 0;
  }
  let end = Math.min(start + 140, lineLength);
  if (end > lineLength - 5) {
    end = lineLength;
  }
  if (end === lineLength) {
    start = Math.max(end - 140, 0);
  }
  newLine = newLine.slice(start, end);
  if (start > 0) {
    newLine = `'{snip} ${newLine}`;
  }
  if (end < lineLength) {
    newLine += " {snip}";
  }
  return newLine;
}
function isValidSentryStackFrame(frame) {
  return !!frame.filename && !!frame.lineno && !!frame.colno;
}
function contextLinesHandler(req, res) {
  if (req.method !== "PUT") {
    res.writeHead(405);
    res.end();
    return;
  }
  let requestBody = "";
  req.on("data", (chunk) => {
    requestBody += chunk;
  });
  req.on("end", async () => {
    const stacktrace = parseStackTrace(requestBody);
    if (!stacktrace) {
      res.writeHead(500);
      res.end();
      return;
    }
    for (const frame of stacktrace.frames ?? []) {
      if (!isValidSentryStackFrame(frame) || // let's ignore dependencies for now with this naive check
      frame.filename.includes("/node_modules/")) {
        continue;
      }
      const { filename } = frame;
      if (filename.includes("://")) {
        const generatedCode = await getGeneratedCodeFromServer(frame.filename);
        if (!generatedCode) {
          continue;
        }
        const inlineSourceMapMatch = generatedCode.match(/\/\/# sourceMappingURL=data:application\/json;base64,(.*)/);
        if (inlineSourceMapMatch && inlineSourceMapMatch[1]) {
          const sourceMapBase64 = inlineSourceMapMatch[1];
          const sourceMapContent = Buffer.from(sourceMapBase64, "base64").toString("utf-8");
          await applySourceContextToFrame(sourceMapContent, frame);
        }
      } else if (!filename.includes(":")) {
        try {
          const lines = readFileSync(filename, { encoding: "utf-8" }).split(/\r?\n/);
          addContextLinesToFrame(lines, frame);
        } catch (err) {
          if (err.code !== "ENOENT") {
            throw err;
          }
        }
      }
    }
    const responseJson = JSON.stringify(stacktrace);
    res.writeHead(200, { "Content-Type": "application/json" });
    res.end(responseJson);
  });
}
const defaultLogger = {
  info: (message) => console.log("🔎 [Spotlight]", message),
  warn: (message) => console.warn("🔎 [Spotlight]", message),
  error: (message) => console.error("🔎 [Spotlight]", message),
  debug: (message) => debugEnabled && console.debug("🔎 [Spotlight]", message)
};
let injectedLogger = void 0;
let debugEnabled = false;
function activateLogger(logger2) {
  injectedLogger = logger2;
}
function enableDebugLogging(debug) {
  debugEnabled = debug;
}
const logger = {
  info: (message) => (injectedLogger || defaultLogger).info(message),
  warn: (message) => (injectedLogger || defaultLogger).warn(message),
  error: (message) => (injectedLogger || defaultLogger).error(message),
  debug: (message) => (injectedLogger || defaultLogger).debug(message)
};
class MessageBuffer {
  constructor(size = 100) {
    __publicField(this, "size");
    __publicField(this, "items");
    __publicField(this, "writePos", 0);
    __publicField(this, "head", 0);
    __publicField(this, "timeout", 10);
    __publicField(this, "readers", /* @__PURE__ */ new Map());
    this.size = size;
    this.items = new Array(size);
  }
  put(item) {
    const curTime = (/* @__PURE__ */ new Date()).getTime();
    this.items[this.writePos % this.size] = [curTime, item];
    this.writePos += 1;
    if (this.head === this.writePos) {
      this.head += 1;
    }
    const minTime = curTime - this.timeout * 1e3;
    let atItem;
    while (this.head < this.writePos) {
      atItem = this.items[this.head % this.size];
      if (atItem === void 0)
        break;
      if (atItem[0] > minTime)
        break;
      this.head += 1;
    }
  }
  subscribe(callback) {
    const readerId = generateUuidv4();
    this.readers.set(readerId, callback);
    setTimeout(() => this.stream(readerId));
    return readerId;
  }
  unsubscribe(readerId) {
    this.readers.delete(readerId);
  }
  stream(readerId, readPos = this.head) {
    const cb = this.readers.get(readerId);
    if (!cb)
      return;
    let atReadPos = readPos;
    let item;
    while (true) {
      item = this.items[atReadPos % this.size];
      if (typeof item === "undefined" || atReadPos >= this.writePos) {
        break;
      }
      cb(item[1]);
      atReadPos += 1;
    }
    setTimeout(() => this.stream(readerId, atReadPos), 500);
  }
  clear() {
    this.items = new Array(this.size);
    this.writePos = 0;
    this.head = 0;
    this.readers = /* @__PURE__ */ new Map();
  }
}
function generateUuidv4() {
  let dt = (/* @__PURE__ */ new Date()).getTime();
  return "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g, function(c) {
    let rnd = Math.random() * 16;
    rnd = (dt + rnd) % 16 | 0;
    dt = Math.floor(dt / 16);
    return (c === "x" ? rnd : rnd & 3 | 8).toString(16);
  });
}
function getCorsHeader() {
  return {
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Credentials": "true",
    "Access-Control-Allow-Headers": "*",
    "Access-Control-Allow-Methods": "GET,POST,PUT,OPTIONS,DELETE,PATCH"
  };
}
function enableCORS(handler) {
  return function corsMiddleware(req, res, pathname, searchParams) {
    const headers = {
      ...getCorsHeader(),
      ...getSpotlightHeader()
    };
    for (const [header, value] of Object.entries(headers)) {
      res.setHeader(header, value);
    }
    if (req.method === "OPTIONS") {
      res.writeHead(204, {
        "Cache-Control": "no-cache"
      });
      res.end();
      return;
    }
    return handler(req, res, pathname, searchParams);
  };
}
function getSpotlightHeader() {
  return {
    "X-Powered-by": SERVER_IDENTIFIER
  };
}
function streamRequestHandler(buffer2, incomingPayload) {
  return function handleStreamRequest(req, res, pathname, searchParams) {
    if (req.method === "GET" && req.headers.accept && req.headers.accept == "text/event-stream" && pathname === "/stream") {
      res.writeHead(200, {
        "Content-Type": "text/event-stream",
        "Cache-Control": "no-cache",
        Connection: "keep-alive"
      });
      res.flushHeaders();
      res.write("\n");
      const sub = buffer2.subscribe(([payloadType, data]) => {
        logger.debug(`🕊️ sending to Spotlight`);
        res.write(`event:${payloadType}
`);
        data.split("\n").forEach((line) => {
          res.write(`data:${line}
`);
        });
        res.write("\n");
      });
      req.on("close", () => {
        buffer2.unsubscribe(sub);
        res.end();
      });
    } else if (req.method === "POST") {
      logger.debug(`📩 Received event`);
      let body = "";
      let stream = req;
      const encoding = req.headers["content-encoding"];
      if (encoding === "gzip") {
        stream = req.pipe(createGunzip());
      } else if (encoding === "deflate") {
        stream = req.pipe(createInflate());
      }
      stream.on("readable", () => {
        let chunk;
        while ((chunk = stream.read()) !== null) {
          body += chunk;
        }
      });
      stream.on("end", () => {
        var _a, _b;
        let contentType = (_a = req.headers["content-type"]) == null ? void 0 : _a.split(";")[0].toLocaleLowerCase();
        if (((_b = searchParams == null ? void 0 : searchParams.get("sentry_client")) == null ? void 0 : _b.startsWith("sentry.javascript.browser")) && req.headers.origin) {
          contentType = "application/x-sentry-envelope";
        }
        if (!contentType) {
          logger.warn("No content type, skipping payload...");
        } else {
          buffer2.put([contentType, body]);
        }
        if (process.env.SPOTLIGHT_CAPTURE || incomingPayload) {
          const timestamp = (/* @__PURE__ */ new Date()).getTime();
          const filename = `${(contentType == null ? void 0 : contentType.replace(/[^a-z0-9]/gi, "_")) || "no_content_type"}-${timestamp}.txt`;
          if (incomingPayload) {
            incomingPayload(body);
          } else {
            createWriteStream(filename).write(body);
            logger.info(`🗃️ Saved data to ${filename}`);
          }
        }
        res.writeHead(200, {
          "Cache-Control": "no-cache",
          Connection: "keep-alive"
        });
        res.end();
      });
    } else {
      error405(req, res);
      return;
    }
  };
}
function fileServer(basePath) {
  return function serveFile(req, res, pathname) {
    let filePath = `.${pathname || req.url}`;
    if (filePath === "./") {
      filePath = "./src/index.html";
    }
    const extName = extname(filePath);
    let contentType = "text/html";
    switch (extName) {
      case ".js":
        contentType = "text/javascript";
        break;
      case ".css":
        contentType = "text/css";
        break;
      case ".json":
        contentType = "application/json";
        break;
    }
    readFile(join(basePath, filePath), (error, content) => {
      if (error) {
        return error404(req, res);
      }
      res.writeHead(200, { "Content-Type": contentType });
      res.end(content, "utf-8");
    });
  };
}
function handleHealthRequest(_req, res) {
  res.writeHead(200, {
    "Content-Type": "text/plain",
    ...getCorsHeader(),
    ...getSpotlightHeader()
  });
  res.end("OK");
}
function handleClearRequest(req, res) {
  if (req.method === "DELETE") {
    res.writeHead(200, {
      "Content-Type": "text/plain"
    });
    clearBuffer();
    res.end("Cleared");
  } else {
    error405(req, res);
  }
}
function openRequestHandler(basePath = process.cwd()) {
  return (req, res) => {
    if (req.method !== "POST") {
      res.writeHead(405);
      res.end();
      return;
    }
    let requestBody = "";
    req.on("data", (chunk) => {
      requestBody += chunk;
    });
    req.on("end", () => {
      const targetPath = resolve(basePath, requestBody);
      launchEditor(
        // filename:line:column
        // both line and column are optional
        targetPath,
        // callback if failed to launch (optional)
        (fileName, errorMsg) => {
          logger.error(`Failed to launch editor for ${fileName}: ${errorMsg}`);
        }
      );
      res.writeHead(204);
      res.end();
    });
  };
}
function errorResponse(code) {
  return (_req, res) => {
    res.writeHead(code);
    res.end();
  };
}
const error404 = errorResponse(404);
const error405 = errorResponse(405);
function startServer(buffer2, port, basePath, incomingPayload) {
  const ROUTES = [
    [/^\/health$/, handleHealthRequest],
    [/^\/clear$/, enableCORS(handleClearRequest)],
    [/^\/stream$|^\/api\/\d+\/envelope\/?$/, enableCORS(streamRequestHandler(buffer2, incomingPayload))],
    [/^\/open$/, enableCORS(openRequestHandler(basePath))],
    [RegExp(`^${CONTEXT_LINES_ENDPOINT}$`), enableCORS(contextLinesHandler)],
    [/^.+$/, basePath != null ? enableCORS(fileServer(basePath)) : error404]
  ];
  const server = createServer((req, res) => {
    const url = req.url;
    if (!url) {
      return error404(req, res);
    }
    const { pathname, searchParams } = new URL(url, `http://${req.headers.host || "localhost"}`);
    const route = ROUTES.find((route2) => route2[0].test(pathname));
    if (!route) {
      return error404(req, res);
    }
    return route[1](req, res, pathname, searchParams);
  });
  server.on("error", handleServerError);
  server.listen(port, () => {
    handleServerListen(port, basePath);
  });
  return server;
  function handleServerError(e) {
    if ("code" in e && e.code === "EADDRINUSE") {
      logger.info(`Port ${port} in use, retrying...`);
      setTimeout(() => {
        server.close();
        server.listen(port);
        logger.info(`Port ${port} in use, retrying...`);
      }, 5e3);
    }
  }
  function handleServerListen(port2, basePath2) {
    logger.info(`Sidecar listening on ${port2}`);
    if (basePath2) {
      logger.info(`You can open: http://localhost:${port2} to see the Spotlight overlay directly`);
    }
  }
}
let serverInstance;
const buffer = new MessageBuffer();
const isValidPort = (value) => {
  if (typeof value === "string") {
    const portNumber = Number(value);
    return /^\d+$/.test(value) && portNumber > 0 && portNumber <= 65535;
  }
  return value > 0 && value <= 65535;
};
function isSidecarRunning(port) {
  return new Promise((resolve2) => {
    const options = {
      hostname: "localhost",
      port,
      path: "/health",
      method: "GET",
      timeout: 5e3
    };
    const healthReq = get(options, (res) => {
      const serverIdentifier = res.headers["x-powered-by"];
      if (serverIdentifier === "spotlight-by-sentry") {
        resolve2(true);
      } else {
        resolve2(false);
      }
    });
    healthReq.on("error", () => {
      resolve2(false);
    });
  });
}
function setupSidecar({
  port,
  logger: customLogger,
  basePath,
  debug,
  incomingPayload
} = {}) {
  let sidecarPort = DEFAULT_PORT;
  if (customLogger) {
    activateLogger(customLogger);
  }
  if (debug || process.env.SPOTLIGHT_DEBUG) {
    enableDebugLogging(true);
  }
  if (port && !isValidPort(port)) {
    logger.info("Please provide a valid port.");
    process.exit(1);
  } else if (port) {
    sidecarPort = typeof port === "string" ? Number(port) : port;
  }
  isSidecarRunning(sidecarPort).then((isRunning) => {
    if (isRunning) {
      logger.info(`Sidecar is already running on port ${sidecarPort}`);
    } else {
      if (!serverInstance) {
        serverInstance = startServer(buffer, sidecarPort, basePath, incomingPayload);
      }
    }
  });
}
function clearBuffer() {
  buffer.clear();
}
function shutdown() {
  if (serverInstance) {
    logger.info("Shutting down Server");
    serverInstance.close();
  }
}
process.on("SIGTERM", () => {
  shutdown();
});
export {
  clearBuffer,
  setupSidecar,
  shutdown
};
//# sourceMappingURL=main.js.map
